import dask.array as da
import numpy as np
from matplotlib import pyplot as plt


class UIntHistogram:
    def __init__(self, data=None):
        """
        An unsigned integer histogram class that can be updated with new data.

        :param data: array[uint]
            Optional data of which the initial histogram is built.
        """
        if data is not None:
            assert data.min() >= 0, "Negative data is not supported."
            self.offset, _, self.frequencies = self._get_hist(data)
        else:
            self.offset, self.frequencies = None, None

    @staticmethod
    def _add(list_a, list_b):
        """
        Add two lists element-wise.
        :param list_a: list(numeric)
        :param list_b: list(numeric)
        :return: list(numeric)
        """
        return [e[0] + e[1] for e in zip(list_a, list_b, strict=False)]

    @staticmethod
    def _get_hist(data):
        """
        Compute histogram with integer bins.
        :param data: array[numeric]
        :return: offset, bins, frequencies
        """
        offset = int(data.min())
        bins = int(data.max()) + 1 - offset
        if isinstance(data, da.core.Array):
            freq = (
                da.histogram(data, np.arange(offset, offset + bins + 1))[0]
                .compute()
                .tolist()
            )
        else:
            freq = np.histogram(data, np.arange(offset, offset + bins + 1))[0].tolist()

        return offset, bins, freq

    def _aggregate_histograms(self, offset_data, bins, freq):
        """
        Integrate new frequencies into existing histogram.

        :param offset_data:
        :param bins:
        :param freq:
        :return:
        """
        assert isinstance(freq, list), "freq must be of type List."
        assert isinstance(self.frequencies, list), (
            "frequencies must be of " "type List."
        )
        lower_shift = offset_data - self.offset
        upper_shift = self.offset + self.n_bins() + 1 - (offset_data + bins + 1)

        if lower_shift == 0 and upper_shift == 0:
            # Old and new frequencies cover the same range:
            # [old frequencies]
            # [new frequencies]
            self.frequencies = self._add(self.frequencies, freq)
        elif (
            lower_shift < 0
            and (offset_data + bins >= self.offset)
            and (offset_data + bins <= self.offset + self.n_bins())
        ):
            # New frequencies have additional lower ones.
            #     [old frequencies]
            # [new frequencies]
            frequencies_to = offset_data + bins - self.offset
            freq_from = self.offset - offset_data
            self.frequencies[:frequencies_to] = self._add(
                self.frequencies[:frequencies_to],
                freq[freq_from:],
            )
            self.frequencies = freq[:freq_from] + self.frequencies
            self.offset = offset_data
        elif lower_shift < 0 and (offset_data + bins < self.offset):
            # New frequencies only have additional lower ones.
            #                       [old frequencies]
            # [new frequencies]
            gap_freq = [
                0,
            ] * (self.offset - offset_data - bins)
            self.frequencies = freq + gap_freq + self.frequencies
            self.offset = offset_data
        elif (
            self.offset
            <= offset_data
            <= (self.offset + self.n_bins() - 1)
            < offset_data + bins - 1
        ):
            # New frequencies have additional upper ones.
            #     [old frequencies]
            #            [new frequencies]
            from_frequencies = self.offset + self.n_bins() - offset_data
            to_freq = self.offset + self.n_bins() + 1 - offset_data
            self.frequencies[-from_frequencies:] = self._add(
                self.frequencies[-from_frequencies:],
                freq[:to_freq],
            )
            self.frequencies = self.frequencies + freq[from_frequencies:]
        elif (offset_data) > (self.offset + self.n_bins() - 1):
            # New frequencies have only additional upper ones.
            #     [old frequencies]
            #                           [new frequencies]
            gap_freq = [
                0,
            ] * (offset_data - self.offset - self.n_bins())
            self.frequencies = self.frequencies + gap_freq + freq
        elif lower_shift >= 0 and upper_shift >= 0:
            # New frequencies are completely covered.
            # [          old frequencies          ]
            #           [new frequencies]
            if upper_shift == 0:
                self.frequencies[lower_shift:] = self._add(
                    self.frequencies[lower_shift:], freq
                )
            else:
                self.frequencies[lower_shift:-upper_shift] = self._add(
                    self.frequencies[lower_shift:-upper_shift], freq
                )
        else:
            # Old frequencies are completely covered.
            #     [old frequencies]
            # [    new frequencies    ]
            from_ = self.offset - offset_data
            to = from_ + self.n_bins()
            self.frequencies = self._add(
                self.frequencies,
                freq[from_:to],
            )
            self.frequencies = freq[:from_] + self.frequencies + freq[to:]
            self.offset = offset_data

    def combine(self, histogram):
        """
        Combine this histogram with another UIntHistogram.

        :param histogram: to merge with this
        """
        if self.frequencies is None:
            self.frequencies = histogram.frequencies
            self.offset = histogram.offset
        elif histogram.frequencies is not None:
            self._aggregate_histograms(
                offset_data=histogram.offset,
                bins=histogram.n_bins(),
                freq=histogram.frequencies,
            )

        return self

    def update(self, data):
        """
        Update histogram by adding more data.

        :param data: array(uint)
            Data to be added to the histogram.
        """
        assert data.min() >= 0, "Negative data is not supported."

        if self.frequencies is None:
            self.offset, _, self.frequencies = self._get_hist(data)
        else:
            offset_data, bins, freq = self._get_hist(data)
            self._aggregate_histograms(offset_data=offset_data, bins=bins, freq=freq)

        return self

    def plot(self, width=1):
        """
        Plot histogram.
        """
        if width > 1:
            heights = [
                np.sum(self.frequencies[i : i + width])
                for i in range(self.offset, self.offset + self.n_bins(), width)
            ]
        else:
            heights = self.frequencies

        plt.bar(
            np.arange(self.offset, self.offset + self.n_bins(), width),
            heights,
            width=width,
        )
        plt.show()

    def mean(self):
        """
        Get histogram mean.
        :return: float
        """
        if self.frequencies is None:
            return 0
        return np.sum(
            np.arange(self.offset, self.offset + self.n_bins()) * self.frequencies
        ) / np.sum(self.frequencies)

    def std(self):
        """
        Get histogram standard deviation.
        :return: float
        """
        if self.frequencies is None:
            return 0
        return np.sqrt(
            np.sum(
                (np.arange(self.offset, self.offset + self.n_bins()) - self.mean()) ** 2
                * self.frequencies
            )
            / np.sum(self.frequencies)
        )

    def quantile(self, q):
        """
        Get quantile `q`.
        :param q: uint
        :return: quantile
        """
        assert q >= 0
        assert q <= 1
        if self.frequencies is None:
            return 0
        return self.offset + np.argmax(
            np.cumsum(self.frequencies) / np.sum(self.frequencies) >= q
        )

    def min(self):
        """
        Get minimum.
        :return: uint
        """
        if self.frequencies is None:
            return 0
        return self.offset

    def max(self):
        """
        Get maximum.
        :return: uint
        """
        if self.frequencies is None:
            return 0
        return self.offset + self.n_bins() - 1

    def n_bins(self):
        """Return number of bins.

        :return: uint
        """
        if self.frequencies is None:
            return None
        return len(self.frequencies)

    def save(self, path):
        np.savez(path, frequencies=self.frequencies, offset=self.offset)

    @staticmethod
    def load(path):
        storage = np.load(path, allow_pickle=True)
        hist = UIntHistogram()
        hist.frequencies = storage["frequencies"].tolist()
        hist.offset = storage["offset"]
        return hist



import numpy as np
from scipy.ndimage import gaussian_filter


def create_blurred_illumination_matrix(img, sigma: float = 20):
    """
    Takes an image and blurs it with a Gaussian.

    :param img: The acquired illumination field.
    :param sigma: Gaussian sigma.
    :return: illumination matrix
    """
    matrix = gaussian_filter(input=img.astype(np.float32), sigma=sigma, mode="nearest")
    return matrix



from pathlib import Path

import numpy as np
import pandas as pd
import tifffile
from numpy._typing import NDArray
from tifffile import TiffFile

from faim_ipa.hcs.acquisition import TileAlignmentOptions, WellAcquisition
from faim_ipa.io.metadata import ChannelMetadata
from faim_ipa.stitching.tile import Tile, TilePosition
from faim_ipa.visiview.ome_companion_utils import parse_basic_metadata


class StackedTile(Tile):
    def __init__(
        self,
        path: Path | str,
        shape: tuple[int, int],
        position: TilePosition,
        background_correction_matrix_path: Path | str | None = None,
        illumination_correction_matrix_path: Path | str | None = None,
        *,
        memmap: bool = True,
    ):
        self._memmap = memmap
        super().__init__(
            path=path,
            shape=shape,
            position=position,
            background_correction_matrix_path=background_correction_matrix_path,
            illumination_correction_matrix_path=illumination_correction_matrix_path,
        )

    def load_data(self) -> NDArray:
        data = (
            tifffile.memmap(self.path, mode="r")
            if self._memmap
            else tifffile.imread(self.path)
        )

        data = self._apply_background_correction(data)
        return self._apply_illumination_correction(data)

    def load_data_mask(self) -> NDArray:
        return np.ones(self.shape, dtype=bool)


class RegionAcquisitionSTK(WellAcquisition):
    def __init__(
        self,
        files: pd.DataFrame,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, NDArray] | None,
        illumination_correction_matrices: dict[str, NDArray] | None,
        axes: list[str] | None = None,
        *,
        memmap: bool = True,
    ):
        if axes is None:
            axes = ["c", "z", "y", "x"]
        path = files.iloc[0]["path"]
        with TiffFile(path) as tif:
            metadata = tif.stk_metadata
            if metadata is None:
                msg = f"STK metadata is missing. Please check " f"{path}"
                raise ValueError(msg)
            x_spacing = metadata["XCalibration"]
            y_spacing = metadata["YCalibration"]
            self._yx_spacing = (y_spacing, x_spacing)
            self._z_spacing = metadata["ZDistance"].mean()
            self.tile_shape = tif.asarray().shape

        self._axes = axes
        self._memmap = memmap
        super().__init__(
            files=files,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _assemble_tiles(self) -> list[Tile]:
        tiles = []
        for _i, row in self._files.iterrows():
            file = row["path"]
            time_point = row["time"]
            channel = row["channel"]

            tiles.append(
                StackedTile(
                    path=file,
                    shape=self.tile_shape,
                    position=TilePosition(
                        time=time_point,
                        channel=channel,
                        z=0,
                        y=int(row["Y"] / self._yx_spacing[0]),
                        x=int(row["X"] / self._yx_spacing[1]),
                    ),
                    memmap=self._memmap,
                )
            )

        return tiles

    def get_z_spacing(self) -> float | None:
        return self._z_spacing

    def get_yx_spacing(self) -> tuple[float, float]:
        return self._yx_spacing

    def get_axes(self) -> list[str]:
        return self._axes


class RegionAcquisitionOME(WellAcquisition):
    def __init__(
        self,
        files: pd.DataFrame,
        ome_xml: Path | str,
        alignment: TileAlignmentOptions,
        background_correction_matrices: dict[str, NDArray] | None,
        illumination_correction_matrices: dict[str, NDArray] | None,
        axes: list[str] | None = None,
        *,
        memmap: bool = True,
    ):
        if axes is None:
            axes = ["c", "z", "y", "x"]
        self.metadata = parse_basic_metadata(companion_file=ome_xml)
        self.stage_positions = self.metadata["stage_positions"]
        path = files.iloc[0]["path"]
        with TiffFile(path) as tif:
            self.tile_shape = tif.asarray().shape

        self._axes = axes
        self._memmap = memmap
        super().__init__(
            files=files,
            alignment=alignment,
            background_correction_matrices=background_correction_matrices,
            illumination_correction_matrices=illumination_correction_matrices,
        )

    def _assemble_tiles(self) -> list[Tile]:
        tiles = []
        for _i, row in self._files.iterrows():
            file = row["path"]
            time_point = row["time"]
            channel = row["channel"]

            tiles.append(
                StackedTile(
                    path=file,
                    shape=self.tile_shape,
                    position=TilePosition(
                        time=time_point,
                        channel=channel,
                        z=0,
                        y=int(
                            self.stage_positions[row["well"]][0]
                            / self.metadata["yx_spacing"][0]
                        ),
                        x=int(
                            self.stage_positions[row["well"]][1]
                            / self.metadata["yx_spacing"][1]
                        ),
                    ),
                    memmap=self._memmap,
                )
            )

        return tiles

    def get_z_spacing(self) -> float | None:
        return self.metadata["z_spacing"]

    def get_yx_spacing(self) -> tuple[float, float]:
        return self.metadata["yx_spacing"]

    def get_channels(self) -> list[ChannelMetadata]:
        return self.metadata["channels"]

    def get_axes(self) -> list[str]:
        return self._axes



from pydantic import BaseModel, NonNegativeInt, PositiveFloat


class ChannelMetadata(BaseModel):
    channel_index: NonNegativeInt
    channel_name: str
    display_color: str
    spatial_calibration_x: PositiveFloat
    spatial_calibration_y: PositiveFloat
    spatial_calibration_units: str
    z_spacing: PositiveFloat | None = None
    wavelength: NonNegativeInt | None = None
    exposure_time: PositiveFloat | None = None
    exposure_time_unit: str | None = None
    objective: str
